//   Copyright (c)  2017 John Abbott, Anna M. Bigatti
//   Authors: 2017 Anna M. Bigatti

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


// Source code for MathSAT integration

#include "CoCoA/ExternalLibs-MathSAT.H"

#ifdef CoCoA_WITH_MATHSAT

// #include "CoCoA/BigInt.H"
#include "CoCoA/BigRat.H"
// #include "CoCoA/IntOperations.H"
// #include "CoCoA/TmpPPVector.H"
// #include "CoCoA/convert.H"
#include "CoCoA/error.H"
#include "CoCoA/DenseMatrix.H"
#include "CoCoA/matrix.H"
#include "CoCoA/time.H"
#include "CoCoA/SparsePolyRing.H"
//#include "CoCoA/symbol.H"
#include "CoCoA/RingQQ.H"
#include "CoCoA/verbose.H"

// #include "CoCoA/VectorOperations.H"  // just for debugging

// #include <vector>
// using std::vector;
#include <ostream>
// using std::ostream;
using std::endl;
#include <sstream>  // for 
#include <string>

namespace CoCoA
{

  namespace MathSAT  // "mathsat" is its own namespace
  {


    RelOp ToRelOp(const std::string& s)
    {
      if (s == "eq0") return eq0;
      if (s == "neq0") return neq0;
      if (s == "leq0") return leq0;
      if (s == "lt0") return lt0;
      CoCoA_ERROR("unknown relation "+s, "ToRelOp");
      return eq0; // just to keel the compiler quiet
    }

    //---------------------------------------------------------------
    namespace  // conversion functions, implementation at the end of the file
    {      

      /////////// improve these conversion......
      std::string BRToString(ConstRefRingElem r)
      {
        std::ostringstream buffer;
        buffer << r;
        return buffer.str();
      }

      /////////// improve these conversion......      
      msat_term IndetToMathSAT(msat_env MSE, int var)
      {
        std::ostringstream buf;
        buf << "X" << var;
        std::string name = buf.str();
        msat_type rat = msat_get_rational_type(MSE);
        msat_decl d = msat_declare_function(MSE, name.c_str(), rat);
        msat_term x = msat_make_constant(MSE, d);
        return x;
      }
      
      
      msat_term RingElemToMathSAT(msat_env MSE, ConstRefRingElem L)
      {
        msat_term res = msat_make_number(MSE, "0");
        long v;
        for (SparsePolyIter iter=BeginIter(L); !IsEnded(iter); ++iter)
        {
          msat_term c = msat_make_number(MSE, BRToString(coeff(iter)).c_str());
          if (IsOne(PP(iter)))  res = msat_make_plus(MSE, res, c);
          else
          {
            if (IsIndet(v, PP(iter)))
            {
              msat_term x = IndetToMathSAT(MSE,v);
              res = msat_make_plus(MSE, res, msat_make_times(MSE, c, x));
            }
            else
              CoCoA_ERROR("must be an indeterminate", "EqualToMathSAT");
          }
        }
        return res;
      }
      
      msat_term RowToMathSAT(msat_env MSE, ConstMatrixView M, long i)
      {
        CoCoA_ASSERT_ALWAYS(i>=0 && i<NumRows(M));
        long v = NumCols(M);
        --v;
        // change sign to constant term: 1 2 3 ==> 1*x+2*y 3 <=> 0
        msat_term res = msat_make_number(MSE, BRToString(M(i,v)).c_str());
        for (--v; v>=0; --v)
        {
          msat_term c = msat_make_number(MSE, BRToString(M(i,v)).c_str());
          msat_term x = IndetToMathSAT(MSE,v);
          res = msat_make_plus(MSE, res, msat_make_times(MSE, c, x));
        }
        return res;
      }
      
      int GetCoeffIndetIndex(msat_term x)
      {
        msat_decl d = msat_term_get_decl(x);
        char *name = msat_decl_get_name(d);
        int var = atoi(name+1);
        msat_free(name);
        return var;
      }
    } // end of anonymous namespace
    //---------------------------------------------------------------




    //---------------------------------------------------------------
    // --- env -------------------

    env::env()
    {
      myNumIndetsValue = 0;
      // create a MathSAT environment
      msat_config cfg = msat_create_config();
      msat_set_option(cfg, "model_generation", "true");
      myEnvValue = msat_create_env(cfg);
      msat_destroy_config(cfg);
    }
    

    env::~env()
    { msat_destroy_env(myEnvValue); }

    //--- access to msat_env: polynomial ----------------------------
    void env::myAddEq0(ConstRefRingElem Lin)
    {
      VerboseLog VERBOSE("MathSAT::env::myAddEq0");
      VERBOSE(80) << Lin << " == 0" << endl;
      if (myNumIndetsValue == 0)
        myNumIndetsValue = NumIndets(owner(Lin));
      else
        CoCoA_ASSERT_ALWAYS(myNumIndetsValue == NumIndets(owner(Lin)));
      msat_term L = RingElemToMathSAT(myEnvValue, Lin);
      msat_term Z = msat_make_number(myEnvValue,"0");
      msat_assert_formula(myEnvValue, msat_make_equal(myEnvValue, L,Z));
    }

    void env::myAddNeq0(ConstRefRingElem Lin)
    {
      VerboseLog VERBOSE("MathSAT::env::myAddNeq0");
      VERBOSE(80) << Lin << " != 0" << endl;
      if (myNumIndetsValue == 0)
        myNumIndetsValue = NumIndets(owner(Lin));
      else
        CoCoA_ASSERT_ALWAYS(myNumIndetsValue == NumIndets(owner(Lin)));
      msat_term L = RingElemToMathSAT(myEnvValue, Lin);
      msat_term Z = msat_make_number(myEnvValue,"0");
      msat_assert_formula(myEnvValue, msat_make_not(myEnvValue, msat_make_equal(myEnvValue, L,Z)));
    }
    

    void env::myAddLeq0(ConstRefRingElem Lin)
    {
      VerboseLog VERBOSE("MathSAT::env::myAddLeq0");
      VERBOSE(80) << Lin << " <= 0" << endl;
      if (myNumIndetsValue == 0)
        myNumIndetsValue = NumIndets(owner(Lin));
      else
        CoCoA_ASSERT_ALWAYS(myNumIndetsValue == NumIndets(owner(Lin)));
      msat_term L = RingElemToMathSAT(myEnvValue, Lin);
      msat_term Z = msat_make_number(myEnvValue,"0");
      msat_assert_formula(myEnvValue, msat_make_leq(myEnvValue, L, Z));
    }
    
      
    void env::myAddLt0(ConstRefRingElem Lin)
    {
      VerboseLog VERBOSE("MathSAT::env::myAddLt0");
      VERBOSE(80) << Lin << " < 0" << endl;
      if (myNumIndetsValue == 0)
        myNumIndetsValue = NumIndets(owner(Lin));
      else
        CoCoA_ASSERT_ALWAYS(myNumIndetsValue == NumIndets(owner(Lin)));
      msat_term L = RingElemToMathSAT(myEnvValue, Lin);
      msat_term Z = msat_make_number(myEnvValue,"0");
      msat_assert_formula(myEnvValue, msat_make_not(myEnvValue, msat_make_leq(myEnvValue, Z,L)));
    }

    //--- access to msat_env: matrix ------------------------------
    void env::myAddEq0(ConstMatrixView M)
    {
      VerboseLog VERBOSE("MathSAT::env::myAddEq0");
      VERBOSE(80) << M << endl;
      if (myNumIndetsValue == 0)
        myNumIndetsValue = NumCols(M)-1;
      else
        CoCoA_ASSERT_ALWAYS(myNumIndetsValue == NumCols(M)-1);
      msat_term Z = msat_make_number(myEnvValue,"0");
      double t0=CpuTime();
      double t1, st1=0;
      for (long i=0; i<NumRows(M); ++i)
      {
        t1 = CpuTime();
        msat_term L = RowToMathSAT(myEnvValue, M, i);
        st1 += (CpuTime()-t1);
        msat_assert_formula(myEnvValue, msat_make_equal(myEnvValue, L,Z));
      }
      //      VERBOSE(90) << "RowToMathSAT   time = " << st1 << std::endl;
      VERBOSE(90) << "total for-loop time = " << CpuTime()-t0 << std::endl;
    }
    
      
    void env::myAddNeq0(ConstMatrixView M)
    {
      VerboseLog VERBOSE("MathSAT::env::myAddNeq0");
      if (myNumIndetsValue == 0)
        myNumIndetsValue = NumCols(M)-1;
      else
        CoCoA_ASSERT_ALWAYS(myNumIndetsValue == NumCols(M)-1);
      VERBOSE(80) << M << endl;
      msat_term Z = msat_make_number(myEnvValue,"0");
      double t0=CpuTime();
      for (long i=0; i<NumRows(M); ++i)
      {
        msat_term L = RowToMathSAT(myEnvValue, M, i);
        msat_assert_formula(myEnvValue, msat_make_not(myEnvValue, msat_make_equal(myEnvValue, L,Z)));
      }
      VERBOSE(90) << "total for-loop time = " << CpuTime()-t0 << std::endl;
    }
    

    void env::myAddLeq0(ConstMatrixView M)
    {
      VerboseLog VERBOSE("MathSAT::env::myAddLeq0");
      if (myNumIndetsValue == 0)
        myNumIndetsValue = NumCols(M)-1;
      else
        CoCoA_ASSERT_ALWAYS(myNumIndetsValue == NumCols(M)-1);
      VERBOSE(80) << M << endl;
      msat_term Z = msat_make_number(myEnvValue,"0");
      double t0=CpuTime();
      for (long i=0; i<NumRows(M); ++i)
      {
        msat_term L = RowToMathSAT(myEnvValue, M, i);
        msat_assert_formula(myEnvValue, msat_make_leq(myEnvValue, L, Z));
      }
      VERBOSE(90) << "total for-loop time = " << CpuTime()-t0 << std::endl;
    }
    
      
    void env::myAddLt0(ConstMatrixView M)
    {
      VerboseLog VERBOSE("MathSAT::env::myAddLt0");
      if (myNumIndetsValue == 0)
        myNumIndetsValue = NumCols(M)-1;
      else
        CoCoA_ASSERT_ALWAYS(myNumIndetsValue == NumCols(M)-1);
      VERBOSE(80) << M << endl;
      msat_term Z = msat_make_number(myEnvValue,"0");
      double t0=CpuTime();
      for (long i=0; i<NumRows(M); ++i)
      {
        msat_term L = RowToMathSAT(myEnvValue, M, i);
        msat_assert_formula(myEnvValue, msat_make_not(myEnvValue, msat_make_leq(myEnvValue, Z,L)));
      }
      VERBOSE(90) << "total for-loop time = " << CpuTime()-t0 << std::endl;
    }



    //--- access to msat_env: LinSolve ------------------------------
    matrix env::myLinSolve() const
    {
      VerboseLog VERBOSE("MathSAT::env::myLinSolve");
      double t0=CpuTime();
      msat_solve(myEnvValue);
      t0 -= CpuTime();
      VERBOSE(90) << "msat_solve time = " << -t0 << std::endl;
      if (msat_solve(myEnvValue) == MSAT_UNSAT)
        return NewDenseMat(RingQQ(), 0,0);
      matrix M = NewDenseMat(RingQQ(), myNumIndetsValue, 1);
      BigRat ans;
      msat_model_iterator it = msat_create_model_iterator(myEnvValue);
      while (msat_model_iterator_has_next(it)) {
        msat_term t, v;
        msat_model_iterator_next(it, &t, &v);
        msat_term_to_number(myEnvValue, v, mpqref(ans));
        SetEntry(M,  MathSAT::GetCoeffIndetIndex(t),0,   ans);
      }
      return M;
    }
    

    bool env::IamSatisfiable() const
    {
      return msat_solve(myEnvValue) == MSAT_SAT;
    }
    

    void AddConstraint(env& E, RelOp rel, ConstMatrixView M)
    {
      switch (rel)
      {
      case eq0:  E.myAddEq0(M); break;
      case neq0: E.myAddNeq0(M); break;
      case leq0: E.myAddLeq0(M); break;
      case lt0:  E.myAddLt0(M); break;
      }
    }    

    void AddConstraint(env& E, RelOp rel, ConstRefRingElem Lin)
    {
      switch (rel)
      {
      case eq0:  E.myAddEq0(Lin); break;
      case neq0: E.myAddNeq0(Lin); break;
      case leq0: E.myAddLeq0(Lin); break;
      case lt0:  E.myAddLt0(Lin); break;
      }
    }    

    void AddEq0(env& E, ConstRefRingElem Lin)  { E.myAddEq0(Lin); }
    void AddNeq0(env& E, ConstRefRingElem Lin) { E.myAddNeq0(Lin); }
    void AddLeq0(env& E, ConstRefRingElem Lin) { E.myAddLeq0(Lin); }
    void AddLt0(env& E, ConstRefRingElem Lin)  { E.myAddLt0(Lin); }

    void AddEq0(env& E, ConstMatrixView M)  { E.myAddEq0(M); }
    void AddNeq0(env& E, ConstMatrixView M) { E.myAddNeq0(M); }
    void AddLeq0(env& E, ConstMatrixView M) { E.myAddLeq0(M); }
    void AddLt0(env& E, ConstMatrixView M)  { E.myAddLt0(M); }




    matrix LinSolve(const env& E)
    { return E.myLinSolve(); }
    
    //---------------------------------------------------------------
    

  } // namespace MathSAT
} // namespace CoCoA

// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/ExternalLibs-MathSAT.C,v 1.7 2017/08/08 14:25:48 bigatti Exp $
// $Log: ExternalLibs-MathSAT.C,v $
// Revision 1.7  2017/08/08 14:25:48  bigatti
// -- updated for SC2
//
// Revision 1.6  2017/07/24 14:52:55  bigatti
// -- cleaned up design for MathSAT wrapper class
//
// Revision 1.5  2017/07/14 09:31:15  bigatti
// -- new class MathSAT::env.  Consequent changes
//
// Revision 1.4  2017/07/12 16:44:53  bigatti
// -- developed experimental code for MathSat and moved it from example
//    to ExternalLib-MathSAT.[CH]
//
// Revision 1.3  2017/04/27 16:18:50  bigatti
// -- changed copyright
//
// Revision 1.2  2017/03/13 12:30:37  abbott
// Summary: Move #ifdef guard to after inclusion of header file (so that defns from PREPROCESSOR_DEFNS.H are visible)
//
// Revision 1.1  2017/02/24 08:20:31  bigatti
// -- first import
//

#endif
