//   Copyright (c)  1997-2007,2016  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.

#include "CoCoA/SmallFpImpl.H"
#include "CoCoA/matrix.H"
#include "CoCoA/DenseMatrix.H"
#include "CoCoA/convert.H"

#include <vector>

// /* Two bits used in the return value */
// const int FFsolve_new_pivot = 2;
// const int FFsolve_invalid = 1;



/* Find a solution to a system of linear equations (in place).          */
/* Several right hand side vectors may be specified.                    */
/* Method is just direct Gaussian elimination -- fine for finite fields */
/* The value returned is the rank of M.                                 */
/* The result is placed in soln; if there is no solution for some rhs   */
/* vector then its solution has first coordinate equal to p, the modulus.*/

namespace CoCoA
{

class MatrixFp
{
public:
  MatrixFp(int nrows, int ncols);
//  MatrixFp(const SmallFpImpl& Modp, const MatrixFpNonRed& M);
  friend long NumRows(const MatrixFp& M);
  friend long NumCols(const MatrixFp& M);
  SmallFpImpl::value& operator()(long r, long c);
  const SmallFpImpl::value& operator()(long r, long c) const;
//  friend void SetEntry(MatrixFp& M, long r, long c, SmallFpImpl::value x);
private:
  long myRows;
  long myCols;
  std::vector< std::vector< SmallFpImpl::value > > myM;
};

  MatrixFp::MatrixFp(int nrows, int ncols):
      myRows(nrows),
      myCols(ncols),
      myM(nrows, std::vector<SmallFpImpl::value>(ncols))
  {}


  long NumRows(const MatrixFp& M) { return M.myRows; }
  long NumCols(const MatrixFp& M) { return M.myCols; }
  inline SmallFpImpl::value& MatrixFp::operator()(long r, long c) { return myM[r][c]; }
  inline const SmallFpImpl::value& MatrixFp::operator()(long r, long c) const { return myM[r][c]; }

class MatrixFpNonRed
{
public:
  MatrixFpNonRed(long nrows, long ncols);
//  MatrixFpNonRed();
  void mySwapRows(long r1, long r2) {CoCoA_ASSERT(0 <= r1 && r1 < myRows && 0 <= r2 && r2 < myRows); if (r1 != r2) std::swap(myM[r1],myM[r2]);} ;
  SmallFpImpl::NonRedValue& operator()(long r, long c);
  const SmallFpImpl::NonRedValue& operator()(long r, long c) const;
//  friend void SetEntry(MatrixFp& M, long r, long c, SmallFpImpl::value x);
  
private:
  long myRows;
  long myCols;
  std::vector< std::vector< SmallFpImpl::NonRedValue > > myM;
};

MatrixFpNonRed::MatrixFpNonRed(long nrows, long ncols):
    myRows(nrows),
    myCols(ncols),
    myM(nrows, std::vector<SmallFpImpl::NonRedValue>(ncols))
{}

  inline SmallFpImpl::NonRedValue& MatrixFpNonRed::operator()(long r, long c) { return myM[r][c]; }
  inline const SmallFpImpl::NonRedValue& MatrixFpNonRed::operator()(long r, long c) const { return myM[r][c]; }


class LinSolver
{
public:
  LinSolver(const SmallFpImpl& Modp, const MatrixFp& M, const MatrixFp& rhs);
//??  LinSolver(const matrix& M, const matrix& rhs);
  friend long rk(LinSolver& LS);
  friend const MatrixFp& soln(/*const*/ LinSolver& LS);
private:
  void SolveByGauss();
private:
  long myNrows;
  long myNcols;
  long myRHScols;
  long myRank; // -1 means not yet solved, non-neg is true value
  const SmallFpImpl& myModp;
  MatrixFpNonRed myM; // matrix and rhs concatenated
  MatrixFp mySoln;
  std::vector<bool> mySolnIsValid;
  std::vector<long> myPivotRow;
};

 long rk(LinSolver& LS) { if (LS.myRank < 0) LS.SolveByGauss(); return LS.myRank; }
const MatrixFp& soln(/*const*/ LinSolver& LS) { if (LS.myRank < 0) LS.SolveByGauss(); return LS.mySoln;} 

LinSolver::LinSolver(const SmallFpImpl& Modp, const MatrixFp& M, const MatrixFp& rhs):
    myNrows(NumRows(M)),
    myNcols(NumCols(M)),
    myRHScols(NumCols(rhs)),
    myRank(-1),
    myModp(Modp),
    myM(myNrows, myNcols+myRHScols),
    mySoln(myNrows, myRHScols),
    mySolnIsValid(myRHScols), /// ??? init val
    myPivotRow(myNcols)  ///??? init val?
{
  CoCoA_ASSERT(NumRows(rhs) == myNrows);
//  CoCoA_ASSERT(NumRows(rhs) == myNrows);
  // Fill myM
  for (int i=0; i < myNrows; ++i)
  {
    for (int j=0; j < myNcols; ++j)
    {
      myM(i,j) = M(i,j);
    }
    for (int j=0; j < myRHScols; ++j)
    {
      myM(i, j+myNcols) = rhs(i,j);
    }
  }
}

//int FFsolve(FFmat soln, int *shape, FFmat matrix, FFmat RHS)
//int LinSolveInPlace(vector<vector< SmallFpImpl::NonRedValue> >& M, int ncols)
void LinSolver::SolveByGauss()//const SmallFpImpl& Modp, MatrixFp& soln, const MatrixFp& M, const MatrixFp& rhs)
{
  int ReducedRows = 0;
  const int AllCols = myNcols+myRHScols;
//  trows = 0;
  const SmallFpImpl& Modp = myModp;
  const long TimeToReduce = Modp.myMaxIters();
  long count = 0;
//  vector<int> PivotRow(ncols);
//  pivot_row = (int*)MALLOC(ncols*sizeof(int));
  // Do Gauss redn col-by-col from lhs
  for (int i=0; i < myNcols; ++i)
  {
    bool ZeroCol = true;
    SmallFpImpl::value pivot;
    for (int j=ReducedRows; j < myNrows; ++j)
    {
      pivot = Modp.myNormalize(myM(j,i));
      if (!IsZero(pivot)) { ZeroCol = false; break;}
    }
    if (ZeroCol)
    {
///      if (shape[i]) { FREE(pivot_row); return FFsolve_invalid; }
      continue;
    }
// What does next block do????
    // if (!shape[i])
    // {
    //   shape[i] = 1;
    //   result |= FFsolve_new_pivot;
    //   for (k=i+1; k < ncols; ++k) shape[k] = 0;
    // }
    myPivotRow[ReducedRows] = i;
    myM.mySwapRows(ReducedRows, i);
    // swap = M[j]; M[j] = M[trows]; M[trows] = swap;
    // swap = rhs[j]; rhs[j] = rhs[trows]; rhs[trows] = swap;
//    const SmallFpImpl::value invMii = Modp.myRecip(myM(ReducedRows,i));
    const SmallFpImpl::value inv_pivot = Modp.myRecip(pivot);
    for (int j=i; j < AllCols; ++j)
      myM(ReducedRows,j) = Modp.myMul(inv_pivot, Modp.myNormalize(myM(ReducedRows,j)));
//    for (j=i; j < ncols; j++) M[trows][j] = FFmul(M[trows][j]%p, invMii);
//    for (j=0; j < r; j++) rhs[trows][j] = FFmul(rhs[trows][j]%p, invMii);

    // Now zero the whole column (except the pivot, of course!)
    for (int j=0; j < myNrows; j++)
    {
      if (j == ReducedRows) continue;
      SmallFpImpl::value Mji = Modp.myNormalize(myM(j,i));
      if (IsZero(Mji)) continue;
      Mji = Modp.myNegate(Mji);
      for (int k=i+1; k < AllCols; ++k)
        myM(j,k) += Mji*myM(ReducedRows,k);
      // for (int k=0; k < r; k++) rhs[j][k] += Mji*rhs[trows][k];
      // for (k=i+1; k < ncols; k++) M[j][k] += Mji*M[trows][k];
      myM(j,i) = zero(SmallFp);
    }
    ++ReducedRows;
    if (++count < TimeToReduce) continue;
    count = 0;
    for (int j=0; j < myNrows; ++j)
    {
      for (int k=i+1; k < AllCols; ++k)
        myM(j,k) = Modp.myHalfNormalize(myM(j,k));
      // for (k=0; k < r; k++)
      //   if (rhs[j][k] >= shift) rhs[j][k] -= shift;
      // for (k=i+1; k < ncols; k++)
      //   if (M[j][k] >= shift) M[j][k] -= shift;
    }
  }

  for (int j=0; j < myRHScols; ++j)
  {
    /* If any coord beyond ReducedRows is non-zero, no solution exists. */
    for (int i=ReducedRows; i < myNrows; ++i)
    {
      myM(i, j+myNcols) = Modp.myNormalize(myM(i,j));
      if (!IsZero(Modp.myNormalize(myM(i,j)))) { mySolnIsValid[j] = false; break; }
    }
    // There is a soln, so put it into mySoln:
    int k = ReducedRows-1;
    for (int i=myNcols-1; i >= 0; --i)
    {
      if (k >= 0 && i == myPivotRow[k])
      { mySoln(i,j) = Modp.myNormalize(myM(k, j + myNcols)); --k; }
//default value!!      else mySoln(i,j) = 0;
    }
  }

}


  matrix LinSolveFp(ConstMatrixView M, ConstMatrixView rhs)
  {
    ring Fp = RingOf(M);
    // assume RingOf(rhs) == Fp
    const int p = ConvertTo<long>(characteristic(Fp));
    SmallFpImpl Modp(p);

    const long rows = NumRows(M);
    const long cols = NumCols(M);
    const long rhs_cols  = NumCols(rhs);
    MatrixFp Mcopy(rows, cols);
    MatrixFp RHScopy(rows, rhs_cols);
    for (int i=0; i < rows; ++i)
      for (int j=0; j < cols; ++j)
        Mcopy(i,j) = Modp.myReduce(ConvertTo<long>(M(i,j)));
    for (int i=0; i < rows; ++i)
      for (int j=0; j < rhs_cols; ++j)
        RHScopy(i,j) = Modp.myReduce(ConvertTo<long>(rhs(i,j)));
    LinSolver LS(Modp, Mcopy, RHScopy);
    const MatrixFp& Msoln = soln(LS);
    matrix ans = NewDenseMat(Fp, rows, rhs_cols);
    for (int i=0; i < rows; ++i)
      for (int j=0; j < rhs_cols; ++j)
        SetEntry(ans,i,j,  Modp.myExportNonNeg(Msoln(i,j)));
    return ans;
  }
  

} // end of namespace CoCoA
